#!/usr/bin/env python

import argparse
import json
import pkg_resources
import re


def gather_intel(app):
    """ Get scripts and entry points info from app """
    dist = pkg_resources.get_distribution(app)
    intel = {
        'scripts': get_scripts(dist),
        'group_specs': get_group_specs(dist),
    }

    return intel


def get_scripts(dist):
    console_scripts = dist.get_entry_map('console_scripts')

    if console_scripts:
        return list(console_scripts.keys())

    scripts = set()

    for record_file in ['RECORD', 'installed-files.txt', 'SOURCES.txt']:
        try:
            records = dist.get_metadata(record_file)

        except Exception:
            continue

        if records:
            bin_re = re.compile('\.\./bin/([^,]+),?')

            for line in records.split('\n'):
                match = bin_re.search(line)

                if match:
                    scripts.add(match.group(1))

    return list(scripts)


def get_group_specs(dist):
    app_specs = []

    for app, spec in dist.get_entry_map('autopip').items():
        app_specs.append((app, spec.module_name, next(iter(spec.extras), None)))

    return app_specs


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=gather_intel.__doc__)
    parser.add_argument('app', help='App to get info from')
    args = parser.parse_args()

    print(json.dumps(gather_intel(args.app)))



#!/usr/bin/env python

import argparse
import os
import platform
import re
import shutil
import subprocess
import sys

SUPPORTED_VERSIONS = ('3.6', '3.7')
IS_DEBIAN = platform.system() == 'Linux' and os.path.exists('/etc/debian_version')
IS_OLD_UBUNTU = (IS_DEBIAN and os.path.exists('/etc/lsb-release') and
                 re.search('RELEASE=1[46]', open('/etc/lsb-release').read()))
IS_MACOS = platform.system() == 'Darwin'
SUDO = 'sudo ' if os.getuid() else ''

parser = argparse.ArgumentParser(description='Check and fix Python installation')
parser.add_argument('--autofix', action='store_true', help='Automatically fix any problems found')
parser.add_argument('--version', default=SUPPORTED_VERSIONS[0], choices=SUPPORTED_VERSIONS,
                    help='Python version to check')
args = parser.parse_args()

PY_VERSION = args.version
AUTOFIX = args.autofix


def check_sudo():
    if not run('which sudo', return_output=True):
        error('! sudo is not installed.')
        print('  Please ask an administrator to install it and run this again.')
        sys.exit(1)


def check_apt():
    os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
    run(SUDO + 'apt-get install -y apt-utils', return_output=True)


def check_curl():
    if not run('which curl', return_output=True):
        error('! curl is not installed.')
        if IS_DEBIAN:
            raise AutoFixSuggestion('To install, run', SUDO + 'apt-get install -y curl')
        sys.exit(1)


def check_python():
    py3_path = run('which python' + PY_VERSION, return_output=True)
    if not py3_path:
        error('! Python ' + PY_VERSION + ' is not installed.')
        if '--version' not in sys.argv:
            print('  autopip supports Python {}.'.format(', '.join(SUPPORTED_VERSIONS)) +
                  ' To check a different version, re-run using "python - --version x.y"')

        if IS_OLD_UBUNTU:
            raise AutoFixSuggestion('To install, run',
                                    (SUDO + 'apt-get update',
                                     SUDO + 'apt-get install -y software-properties-common',
                                     SUDO + 'add-apt-repository -y ppa:deadsnakes/ppa',
                                     SUDO + 'apt-get update',
                                     SUDO + 'apt-get install -y python' + PY_VERSION))

        elif IS_DEBIAN:
            raise AutoFixSuggestion('To install, run',
                                    (SUDO + 'apt-get update', SUDO + 'apt-get install -y python' + PY_VERSION))

        elif IS_MACOS:
            raise AutoFixSuggestion('To install, run', 'brew install python')

        print('  Please install Python ' + PY_VERSION +
              ' per http://docs.python-guide.org/en/latest/starting/installation/')
        sys.exit(1)


def check_pip():
    if not run('which pip3', return_output=True):
        error('! pip3 is not installed.')
        if IS_DEBIAN:
            raise AutoFixSuggestion('To install, run', SUDO + 'apt-get install -y python3-pip')

        elif IS_MACOS:
            raise AutoFixSuggestion('To install, run', 'curl -s https://bootstrap.pypa.io/get-pip.py | ' +
                                    SUDO + 'python' + PY_VERSION)
        print('  If your package repo has a *-pip package for Python ' + PY_VERSION +
              ', then installing it from there is recommended.')
        print('  To install directly, run: curl -s https://bootstrap.pypa.io/get-pip.py | ' +
              SUDO + 'python' + PY_VERSION)
        sys.exit(1)

    version_full = run('pip3 --version', return_output=True)

    if 'python ' + PY_VERSION not in version_full:
        print('  ' + version_full.strip())
        error('! pip3 is pointing to another Python version and not Python ' + PY_VERSION)
        if '--version' not in sys.argv:
            print('  autopip supports Python {}.'.format(', '.join(SUPPORTED_VERSIONS)) +
                  ' To check a different version, re-run using "python - --version x.y"')

        raise AutoFixSuggestion('To re-install for Python ' + PY_VERSION + ', run',
                                'curl -s https://bootstrap.pypa.io/get-pip.py | ' + SUDO + 'python' + PY_VERSION)

    version_str = version_full.split()[1]
    version = tuple(map(_int_or, version_str.split('.', 2)))
    if version < (9, 0, 3):
        error('! Version is', version_str + ', but should be 9.0.3+')
        raise AutoFixSuggestion('To upgrade, run', SUDO + 'pip3 install pip==9.0.3')


def check_venv():
    test_venv_path = '/tmp/check-python-venv-{}'.format(os.getpid())

    try:
        try:
            run('python' + PY_VERSION + ' -m venv ' + test_venv_path, stderr=subprocess.STDOUT, return_output=True,
                raises=True)

        except Exception:
            error('! Could not create virtual environment.')
            if IS_DEBIAN:
                raise AutoFixSuggestion('To install, run', SUDO + 'apt-get install -y python' + PY_VERSION + '-venv')
            print('  Please make sure Python venv package is installed.')
            sys.exit(1)

    finally:
        shutil.rmtree(test_venv_path, ignore_errors=True)

    try:
        try:
            run('virtualenv --python python' + PY_VERSION + ' ' + test_venv_path, stderr=subprocess.STDOUT,
                return_output=True,
                raises=True)

        except Exception as e:
            if run('which virtualenv', return_output=True):
                error('! Could not create virtual environment.')
                print('  ' + str(e))
                sys.exit(1)

            else:
                error('! virtualenv is not installed.')
                raise AutoFixSuggestion('To install, run', SUDO + 'pip3 install virtualenv')

    finally:
        shutil.rmtree(test_venv_path, ignore_errors=True)


def check_setuptools():
    try:
        version_str = run('python' + PY_VERSION + ' -m easy_install --version', return_output=True, raises=True)

    except Exception:
        error('! setuptools is not installed.')
        raise AutoFixSuggestion('To install, run', SUDO + 'pip3 install setuptools')

    version_str = version_str.split()[1]
    version = tuple(map(_int_or, version_str.split('.')))
    if version < (39,):
        error('! Version is', version_str + ', but should be 39+')
        raise AutoFixSuggestion('To upgrade, run', SUDO + 'pip3 install -U setuptools')


def check_wheel():
    try:
        version_str = run('python' + PY_VERSION + ' -m wheel version ', return_output=True, raises=True)

    except Exception:
        error('! wheel is not installed.')
        raise AutoFixSuggestion('To install, run', SUDO + 'pip3 install wheel')

    version_str = version_str.split()[1]
    version = tuple(map(_int_or, version_str.split('.')))
    if version < (0, 31):
        error('! Version is', version_str + ', but should be 0.31+')
        raise AutoFixSuggestion('To upgrade, run', SUDO + 'pip3 install -U wheel')


def check_python_dev():
    include_path = run('python' + PY_VERSION +
                       ' -c "from distutils.sysconfig import get_python_inc; print(get_python_inc())"',
                       return_output=True)
    if not include_path:
        error('! Failed to get Python include path, so not sure if Python dev package is installed')
        if IS_DEBIAN:
            raise AutoFixSuggestion('To install, run', SUDO + ' apt-get install -y python' + PY_VERSION + '-dev')
        sys.exit(1)

    python_h = os.path.join(include_path.strip(), 'Python.h')

    if not os.path.exists(python_h):
        error('! Python dev package is not installed as', python_h, 'does not exist')
        if IS_DEBIAN:
            raise AutoFixSuggestion('To install, run', SUDO + 'apt-get install -y python' + PY_VERSION + '-dev')
        sys.exit(1)


def run(cmd, return_output=False, raises=False, **kwargs):
    print('+ ' + str(cmd))

    if '"' in cmd or '|' in cmd:
        kwargs['shell'] = True
    elif isinstance(cmd, str):
        cmd = cmd.split()

    check_call = subprocess.check_output if return_output else subprocess.check_call

    try:
        output = check_call(cmd, **kwargs)

        if isinstance(output, bytes):
            output = output.decode('utf-8')

        return output

    except Exception:
        if return_output and not raises:
            return
        else:
            raise


def _int_or(value):
    try:
        return int(value)
    except Exception as e:
        return value


def error(*msg):
    msg = ' '.join(map(str, msg))
    echo(msg, color=None if AUTOFIX else 'red')


def echo(msg, color=None):
    if sys.stdout.isatty() and color:
        if color == 'red':
            color = '\033[0;31m'
        elif color == 'green':
            color = '\033[92m'

        msg = color + msg + '\033[0m'

    print(msg)


class AutoFixSuggestion(Exception):
    def __init__(self, instruction, cmd):
        super(AutoFixSuggestion, self).__init__(instruction)
        self.cmd = cmd


checks = [check_python, check_pip, check_venv, check_setuptools, check_wheel, check_python_dev]

if AUTOFIX:
    checks.insert(0, check_curl)
    if IS_DEBIAN:
        checks.insert(0, check_apt)
    if SUDO:
        checks.insert(0, check_sudo)

try:
    last_fix = None

    for check in checks:
        print('Checking ' + check.__name__.split('_', 1)[1].replace('_', ' '))

        while True:
            try:
                check()
                break

            except AutoFixSuggestion as e:
                cmds = e.cmd if isinstance(e.cmd, tuple) else (e.cmd,)
                if AUTOFIX:
                    if cmds == last_fix:
                        error('! Failed to fix automatically, so you gotta fix it yourself.')
                        sys.exit(1)

                    else:
                        for cmd in cmds:
                            run(cmd, return_output=True, raises=True)

                        last_fix = cmds

                else:
                    print('  ' + str(e) + ': ' + ' && '.join(cmds) + '\n')
                    print('# Run the above suggested command(s) manually and then re-run to continue checking,')
                    print('  or re-run using "python - --autofix" to run all suggested commands automatically.')
                    sys.exit(1)

        print('')

except Exception as e:
    error('!', str(e))
    sys.exit(1)

except KeyboardInterrupt:
    sys.exit(1)

echo('Python is alive and well. Good job!', color='green')



import argparse
import logging
import signal
import sys

from autopip.constants import UpdateFreq, INSTALL_TIMEOUT_MSG, WAIT_TIMEOUT_MSG, PYTHON_VERSION
from autopip.manager import AppsManager


def main():
    args = cli_args()
    setup_logger(debug=args.debug)
    mgr = AppsManager(debug=args.debug)

    msg = WAIT_TIMEOUT_MSG if args.command == 'update' and args.wait else INSTALL_TIMEOUT_MSG
    signal.signal(signal.SIGALRM, lambda *args, **kwargs: exit(msg))
    signal.alarm(3600)

    try:
        if args.command == 'install':
            mgr.install(args.apps,
                        update=UpdateFreq.from_name(args.update) if args.update else None,
                        python_version=args.python)

        elif args.command == 'list':
            mgr.list(name_filter=args.name_filter, scripts=args.scripts)

        elif args.command == 'update':
            mgr.update(apps=args.apps, wait=args.wait)

        elif args.command == 'uninstall':
            mgr.uninstall(args.apps)

        else:
            raise NotImplementedError('Command {} not implemented yet'.format(args.command))

    except KeyboardInterrupt:
        sys.exit(1)

    except Exception as e:
        if str(e):
            logging.error(f'! {e}', exc_info=args.debug)
        sys.exit(1)


def cli_args():
    """" Get command-line args """
    parser = argparse.ArgumentParser(description='Easily install apps from PyPI and '
                                                 'automatically keep them updated.')
    parser.add_argument('--debug', action='store_true', help='Turn on debug mode')
    subparsers = parser.add_subparsers(title='Commands', help='List of commands', dest='command')

    install_parser = subparsers.add_parser('install',
                                           help='Install apps in their own virtual environments '
                                                'that automatically updates')
    install_parser.add_argument('apps', nargs='+', help='Apps to install')
    install_parser.add_argument('--update', choices=[m.name.lower() for m in UpdateFreq],
                                help='How often to update the app via cron.')
    install_parser.add_argument('--python', metavar='VERSION', default=PYTHON_VERSION,
                                help='Python version to run the app. [default: %(default)s]')

    list_parser = subparsers.add_parser('list', help='List installed apps')
    list_parser.add_argument('name_filter', nargs='?', help='Optionally filter by name')
    list_parser.add_argument('--scripts', action='store_true', help='Show scripts')

    update_parser = subparsers.add_parser('update', help='Update installed apps.')
    update_parser.add_argument('apps', nargs='*', help='Apps to update. Defaults to all apps if run interactively, '
                                                       'otherwise only auto-update enabled apps (e.g. from cron).')
    update_parser.add_argument('--wait', action='store_true', help='Wait for new version to be published '
                                                                   'and then install.')

    uninstall_parser = subparsers.add_parser('uninstall', help='Uninstall apps')
    uninstall_parser.add_argument('apps', nargs='+', help='Apps to uninstall')

    args = parser.parse_args()

    if args.command:
        return args

    parser.print_help()
    sys.exit(0)


def setup_logger(debug=False):
    if debug:
        logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', stream=sys.stdout, level=logging.DEBUG)

    elif sys.stdout.isatty():
        logging.basicConfig(format='%(message)s', stream=sys.stdout, level=logging.INFO)

    else:
        logging.basicConfig(format='%(asctime)s %(message)s', stream=sys.stdout, level=logging.INFO)



from configparser import RawConfigParser
from collections import defaultdict
from functools import lru_cache
import json
from logging import info, error, debug
import os
from pathlib import Path, PurePath
import pkg_resources
import re
import shutil
from subprocess import CalledProcessError, STDOUT
import sys
from time import time, sleep
import urllib.request
import urllib.error

from autopip import crontab, exceptions
from autopip.constants import UpdateFreq, PYTHON_VERSION
from autopip.utils import run, sorted_versions


class AppsManager:
    """ Manages apps """

    def __init__(self, debug=False):
        #: Turn on debug mode
        self.debug = debug

        #: An instance of :cls:`AppsPath`
        self.paths = AppsPath()

        # PyPI url
        self._index_url = None

        # PyPI auth. Tuple of user and password.
        self._index_auth = None

    def install(self, apps, update=None, python_version=None, wait=False):
        """
        Install the given apps

        :param list[str] apps: List of apps to install
        :param UpdateFreq|None update: How often to update
        :param str python_version: Python version to run the app
        :param bool wait: Wait for a new version to be published and then install it.
        """
        self._set_index()

        autopip_path = shutil.which('autopip')
        if (self.paths.is_user and sys.stdout.isatty() and not list(self.apps) and autopip_path
                and autopip_path.startswith(str(self.paths.SYSTEM_SYMLINK_ROOT))):
            info('# Based on permission, this will install to your user home instead of %s',
                 self.paths.SYSTEM_SYMLINK_ROOT)
            info('  To install for everyone, cancel using CTRL+C and then re-run using sudo.')

        failed_apps = []
        printed_wait = False

        for name in apps:
            try:
                if isinstance(name, tuple):  # From app.group_specs()
                    name, update = name
                    if update:
                        update = UpdateFreq.from_name(update)

                app_spec = next(iter(pkg_resources.parse_requirements(name)))
                app, updated = self._install_app(app_spec, update=update, python_version=python_version, wait=wait)

                if updated:
                    printed_wait = False
                    group_specs = app.group_specs()
                    if group_specs:
                        info('This app has defined "autopip" entry points to install: %s', ' '.join(
                             s[0] for s in group_specs))
                        apps.extend([s for s in group_specs if s not in apps])

                elif wait:
                    goback = '\033[1A' if printed_wait else ''
                    print(f'{goback}Waiting for new version of {name} to be published...'.ljust(80))
                    sleep(60)
                    apps.append(name)
                    printed_wait = True

            except Exception as e:
                error(f'! {e}', exc_info=self.debug)
                failed_apps.append(name)
                printed_wait = False

        if failed_apps:
            raise exceptions.FailedAction()

    def _install_app(self, app_spec, update=None, python_version=None, wait=False):
        """ Install the given app """
        app = App(app_spec.name, self.paths, debug=self.debug)
        updated = False

        # Skip update if install was done within the update frequency when run from cron
        if (sys.stdout.isatty() or not app.is_installed or wait
                or update and app.path.stat().st_mtime + update.seconds < time()):
            if app.is_installed:
                app.path.touch()

            version = self._app_version(app_spec)

            if not wait or version != app.current_version:
                updated = app.install(version, app_spec, update=update, python_version=python_version)

        else:
            debug(f'{app.name} does not need to be updated yet.')

        return app, updated

    def _app_version(self, app_spec):
        """ Get app version from PyPI """
        pkg_index_url = self._index_url + app_spec.name + '/'

        try:
            if self._index_auth:
                password_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
                password_mgr.add_password(None, self._index_url, self._index_auth[0], self._index_auth[1])
                handler = urllib.request.HTTPBasicAuthHandler(password_mgr)
                opener = urllib.request.build_opener(handler)

            else:
                opener = urllib.request.build_opener()

            with opener.open(str(pkg_index_url), timeout=10) as fp:
                version_links = fp.read().decode('utf-8')

        except urllib.error.HTTPError as e:
            if e.code == 404:
                raise NameError(f'{app_spec.name} does not exist on {self._index_url}')
            else:
                raise Exception(f'Failed to read from {pkg_index_url}: {e}')

        version_re = re.compile(app_spec.name + r'-(\d+\.\d+\.\d+(?:\.\w+\d+)?)\.')
        versions = []
        matched_versions = []

        for line in version_links.split('\n'):
            match = version_re.search(line)
            if match:
                if match.group(1) in app_spec:
                    matched_versions.append(match.group(1))
                else:
                    versions.append(match.group(1))

        if not matched_versions:
            if versions:
                raise ValueError(f'No app version matching {app_spec} \nAvailable versions: '
                                 + ', '.join(sorted_versions(versions)))
            else:
                raise ValueError(f'No app version found in {pkg_index_url}')

        return sorted_versions(matched_versions)[-1]

    def _set_index(self):
        """ Set PyPI url and auth """
        if not self._index_url:
            for conf_file in ['~/.config/pip/pip.conf', '~/.pip/pip.conf', '/etc/pip.conf']:
                pip_conf = Path(conf_file).expanduser()
                if not pip_conf.exists():
                    continue

                try:
                    parser = RawConfigParser()
                    parser.read(pip_conf)
                    self._index_url = parser.get('global', 'index-url')

                    auth_re = re.compile('//([^:]+)(?::([^@]+))?@')
                    match = auth_re.search(self._index_url)
                    if match:
                        self._index_auth = match.groups()
                        self._index_url = auth_re.sub('//', self._index_url)

                    if not self._index_url.endswith('/'):
                        self._index_url += '/'

                except Exception:
                    pass

            if not self._index_url:
                self._index_url = 'https://pypi.org/simple/'

    @property
    def apps(self):
        """ Iterator for installed apps """
        for app_path in sorted(self.paths.install_root.iterdir()):
            if app_path in {self.paths.symlink_root, self.paths.log_root}:
                continue

            app = App(app_path.name, self.paths)

            if app.is_installed:
                yield app

    def list(self, name_filter=False, scripts=False):
        """
        List installed apps

        :param str name_filter: Filter apps by name
        :param bool scripts: Show scripts
        """
        app_info = []
        info_lens = defaultdict(int)

        for app in self.apps:
            if name_filter and name_filter not in app.name:
                continue

            app_path = str(app.current_path.resolve())

            if app.settings().get('update'):
                update = f"[updates {app.settings()['update']}]"
            else:
                update = ''

            app_info.append((app.name, app.current_version, app_path, update))

            if scripts:
                hide_path = False
                for script in sorted(app.scripts()):
                    script_symlink = self.paths.symlink_root / script
                    if script_symlink.exists() and str(script_symlink.resolve()).startswith(app_path):
                        script_path = str(script_symlink)
                        if hide_path:
                            script_path = script_path.replace(str(script_symlink.parent) + '/',
                                                              ' ' * len(str(script_symlink.parent)) + ' ')
                        app_info.append(('', '', script_path, ''))
                        hide_path = True

        if app_info:
            # Figure out max length of each column
            for info_part in app_info:
                for i, value in enumerate(info_part):
                    info_lens[i] = len(value) if len(value) > info_lens[i] else info_lens[i]

            # Print table
            table_style = '  '.join('{{:{}}}'.format(l) if l else '{}' for l in info_lens.values())
            for info_part in app_info:
                info(table_style.format(*info_part))

        elif name_filter:
            info(f'No apps matching "{name_filter}"')

        else:
            info('No apps are installed yet.')

            autopip_path = shutil.which('autopip')
            if (self.paths.is_user and sys.stdout.isatty() and autopip_path
                    and autopip_path.startswith(str(self.paths.SYSTEM_SYMLINK_ROOT))):
                info('To see apps installed in %s, re-run using sudo.', self.paths.SYSTEM_INSTALL_ROOT)

    def uninstall(self, apps):
        """ Uninstall apps """
        for name in apps:
            if name == 'autopip' and len(list(self.apps)) > 1:
                if apps[-1] == 'autopip':
                    error('! autopip can not be uninstalled until other apps are uninstalled: %s', ' '.join(
                        a.name for a in self.apps if a.name != 'autopip'))
                else:  # Try again after uninstall the other apps
                    apps.append('autopip')

                continue

            app = App(name, self.paths)
            if app.is_installed:
                group_specs = app.group_specs(name_only=True)
                app.uninstall()

                if group_specs:
                    info('This app has defined "autopip" entry points to uninstall: %s', ' '.join(group_specs))
                    apps.extend(group_specs)

            else:
                info(f'{name} is not installed')

        if not list(self.apps):
            try:
                crontab.remove('autopip')
            except Exception as e:
                debug('Could not remove crontab for autopip: %s', e)

    def update(self, apps=None, wait=False):
        """
        Update installed apps

        :param list apps: List of apps to update. Defaults to all.
        :param bool wait: Wait for a new version to be published and then install it.
        """
        app_instances = list([a for a in self.apps if a.name in apps] if apps else self.apps)

        if app_instances:
            app_specs = []
            for app in app_instances:
                settings = app.settings()
                if settings.get('update'):
                    app_specs.append((settings['app_spec'], settings['update']))
                elif sys.stdout.isatty() or wait:
                    app_specs.append((settings.get('app_spec', app.name), None))

            if app_specs:
                self.install(app_specs, wait=wait)

            elif not apps:
                try:
                    crontab.remove('autopip')
                except Exception as e:
                    debug('Could not remove crontab for autopip: %s', e)

        elif list(self.apps):
            info('No apps found matching: %s', ', '.join(apps))
            info('Available apps: %s', ', '.join([a.name for a in self.apps]))

        else:
            info('No apps installed yet.')


class App:
    """ Represents an app that may or may not be installed on disk """

    #: Prefixes of scripts to skip when creating symlinks
    SKIP_SCRIPT_PREFIXES = {'activate', 'pip', 'easy_install', 'python', 'wheel'}

    def __init__(self, name, paths, debug=False):
        """
        :param str name: Name of the app
        :param AppsPath paths: Path paths
        :param bool debug: Turn on debug mode
        """
        self.name = name
        self.paths = paths
        self.debug = debug

        #: Path to install all app versions
        self.path = self.paths.install_root / name

        #: Symlink to current version
        self._current_symlink = self.path / 'current'

        # Unique crontab name to to easily add and remove from crontab
        self._crontab_id = rf'autopip install "{self.name}[^a-z]*"'

    def __repr__(self):
        return f"App('{self.name}')"

    @property
    def is_installed(self):
        """ Is the app installed? """
        return self._current_symlink.exists()

    @property
    def current_path(self):
        """ Path to currently installed app """
        return self._current_symlink if self.is_installed else None

    @property
    def current_version(self):
        """ Currently installed version """
        if self.current_path:
            return self.current_path.resolve().name

    def install(self, version, app_spec, update=None, python_version=None):
        """
        Install the version of the app if it is not already installed

        :param str version: Version of the app to install
        :param pkg_resources.Requirement app_spec: App version requirement from user
        :param UpdateFreq|None update: How often to update. Choose from hourly, daily, weekly, monthly
        :param str python_version: Python version to run app
        :return: True if install or update happened, otherwise False when nothing happened (already installed / non-tty)
        """
        version_path = self.path / version
        prev_version_path = self.current_path and self.current_path.resolve()
        important_paths = [version_path, prev_version_path, self._current_symlink]

        if self.settings():
            if not python_version:
                python_version = self.settings().get('python_version')

            if not update:
                update = self.settings().get('update') and UpdateFreq.from_name(self.settings()['update'])

        if not python_version:
            python_version = PYTHON_VERSION

        if version_path.exists():
            if self.current_version == version:
                # Skip printing / ensuring symlinks / cronjob when running from cron
                if not sys.stdout.isatty():
                    return False

                pinned = str(app_spec).lstrip(self.name)
                pin_info = f' [per spec: {pinned}]' if pinned else ''

                info(f'{self.name} is up-to-date{pin_info}')

            else:
                info(f'{self.name} {version} was previously installed and will be set as the current version')

        else:
            if not shutil.which('python' + python_version):
                error(f'! python{python_version} does not exist. '
                      'Please install it first, or ensure its path is in PATH.')
                sys.exit(1)

            if python_version.startswith('2'):
                venv = f'virtualenv --python=python{python_version}'
            else:
                venv = f'python{python_version} -m venv'

            old_venv_dir = None
            old_path = None
            no_compile = '--no-compile ' if os.getuid() else ''

            info(f'Installing {self.name} to {version_path}')

            os.environ.pop('PYTHONPATH', None)
            if 'VIRTUAL_ENV' in os.environ:
                old_venv_dir = os.environ.pop('VIRTUAL_ENV')
                old_path = os.environ['PATH']
                os.environ['PATH'] = os.pathsep.join([p for p in os.environ['PATH'].split(os.pathsep)
                                                      if os.path.exists(p) and not p.startswith(old_venv_dir)])

            try:
                run(f"""set -e
                    {venv} {version_path}
                    source {version_path / 'bin' / 'activate'}
                    pip install --upgrade pip wheel
                    pip install {no_compile}{self.name}=={version}
                    """, executable='/bin/bash', stderr=STDOUT, shell=True)

            except BaseException as e:
                shutil.rmtree(version_path, ignore_errors=True)

                if isinstance(e, CalledProcessError):
                    if e.output:
                        output = e.output.decode('utf-8')

                        if not python_version.startswith('2'):
                            python2 = shutil.which('python2') or shutil.which('python2.7')
                            if python2:
                                py2_version = Path(python2).name.lstrip('python')
                                if 'is a builtin module since Python 3' in output:
                                    info(f'Failed to install using Python {python_version} venv, '
                                         f'let\'s try using Python {py2_version} virtualenv.')
                                    return self.install(version, app_spec, update=update, python_version=py2_version)

                        info(re.sub(r'(https?://)[^/]+:[^/]+@', r'\1<xxx>:<xxx>@', output))

                    error(f'! Failed to install using Python {python_version}.'
                          ' If this app requires a different Python version, please specify it using --python option.')

                raise

            finally:
                if old_venv_dir:
                    os.environ['VIRTUAL_ENV'] = old_venv_dir
                    os.environ['PATH'] = old_path

            try:
                shutil.rmtree(version_path / 'share' / 'python-wheels', ignore_errors=True)
                run(f"""set -e
                    source {version_path / 'bin' / 'activate'}
                    pip uninstall --yes pip
                    """, executable='/bin/bash', stderr=STDOUT, shell=True)

            except Exception as e:
                debug('Could not remove unnecessary packages/files: %s', e)

        # Update current symlink
        if not self.current_path or self.current_path.resolve() != version_path:
            atomic_symlink = self.path / f'atomic_symlink_for_{self.name}'
            atomic_symlink.symlink_to(version_path)
            atomic_symlink.replace(self._current_symlink)

            # Remove older versions
            for path in [p for p in self.path.iterdir() if p not in important_paths]:
                shutil.rmtree(path, ignore_errors=True)

        current_scripts = self.scripts()

        if not (current_scripts or self.group_specs()):
            self.uninstall()
            raise exceptions.InvalidAction(
                'Odd, there are no scripts included in the app, so there is no point installing it.\n'
                '  autopip is for installing apps with scripts. To install libraries, please use pip.\n'
                '  If you are the app owner, make sure to setup entry_points in setup.py.\n'
                '  See http://setuptools.readthedocs.io/en/latest/setuptools.html#automatic-script-creation')

        self.settings(app_spec=str(app_spec), python_version=python_version)

        # Install cronjobs
        if 'update' not in sys.argv:
            pinning = '==' in str(app_spec) and not str(app_spec).endswith('*')
            if pinning and (self.settings().get('update') or update):
                info('Auto-update will be disabled since we are pinning to a specific version.')
                info('To enable, re-run without pinning to specific version with --update option')

                if self.settings().get('update'):
                    self.settings(update=None)
                    try:
                        crontab.remove(self._crontab_id)
                    except exceptions.MissingError as e:
                        debug('Could not remove crontab for %s: %s', self._crontab_id, e)

            elif update:
                try:
                    autopip_path = shutil.which('autopip')
                    if not autopip_path:
                        raise exceptions.MissingError(
                            'autopip is not available. Please make sure its bin folder is in PATH env var')

                    # Migrate old crontabs
                    try:
                        old_crons = [c for c in crontab.list().split('\n') if c and 'autopip update' not in c]
                        if old_crons:
                            cron_re = re.compile('autopip install "(.+)"')
                            for cron in old_crons:
                                match = cron_re.search(cron)
                                if match:
                                    old_app_spec = next(iter(pkg_resources.parse_requirements(match.group(1))))
                                    old_app = App(old_app_spec.name, self.paths)
                                    if old_app.is_installed:
                                        old_app.settings(app_spec=str(old_app_spec))
                            crontab.remove('autopip')

                    except Exception as e:
                        debug('Could not migrate old crontabs: %s', e)

                    crontab.add(f'{autopip_path} update '
                                f'2>&1 >> {self.paths.log_root / "cron.log"}', cmd_id='autopip update')
                    info(update.name.title() + ' auto-update enabled via cron service')

                    self.settings(update=update.name.lower())

                except Exception as e:
                    error('! Auto-update was not enabled because: %s', e, exc_info=self.debug)

        # Install script symlinks
        prev_scripts = self.scripts(prev_version_path) if prev_version_path else set()
        old_scripts = prev_scripts - current_scripts

        printed_updating = False

        for script in sorted(current_scripts):
            script_symlink = self.paths.symlink_root / script
            script_path = self.current_path / 'bin' / script

            if script_symlink.resolve() == script_path.resolve():
                continue

            if not printed_updating:
                info('Updating script symlinks in {}'.format(self.paths.symlink_root))
                printed_updating = True

            if script_symlink.exists():
                if self.paths.covers(script_symlink) or self.name == 'autopip':
                    atomic_symlink = self.paths.symlink_root / f'atomic_symlink_for_{self.name}'
                    atomic_symlink.symlink_to(script_path)
                    atomic_symlink.replace(script_symlink)
                    info('* {} (updated)'.format(script_symlink.name))

                else:
                    info('! {} (can not change / not managed by autopip)'.format(script_symlink.name))

            else:
                script_symlink.symlink_to(script_path)
                info('+ ' + str(script_symlink.name))

        for script in sorted(old_scripts):
            script_symlink = self.paths.symlink_root / script
            if script_symlink.exists():
                script_symlink.unlink()
                info('- '.format(script_symlink.name))

        if not printed_updating and sys.stdout.isatty() and current_scripts and 'update' not in sys.argv:
            info('Scripts are in {}: {}'.format(self.paths.symlink_root, ', '.join(sorted(current_scripts))))

        # Remove pyc for non-root installs for all versions, not just current.
        if os.getuid():
            try:
                run(f'find {self.path} -name *.pyc | xargs rm', executable='/bin/bash', stderr=STDOUT, shell=True)
            except Exception as e:
                debug('Could not remove *.pyc files: %s', e)

        return True

    def settings(self, **new_settings):
        """ Get or set settings """
        current_settings = {}

        if self.path.exists():
            settings_file = self.path / 'settings.json'
            if settings_file.exists():
                try:
                    current_settings.update(json.load(settings_file.open()))
                except Exception as e:
                    debug('Could not load app settings: %s', e)

            if new_settings:
                current_settings.update(new_settings)
                with settings_file.open('w') as fh:
                    json.dump(current_settings, fh)

        return current_settings

    def scripts(self, path=None):
        """ Set of scripts for the given app path (defaults to current). """
        dist = self._pkg_info(path=path)
        return dist and set(dist['scripts']) or set()

    def group_specs(self, path=None, name_only=False):
        """ List of app specs from this app's "autopip" entry points for the given app path (defaults to current)"""
        app_specs = []
        dist = self._pkg_info(path=path)

        if dist:
            for app, version, update in dist.get('group_specs', []):
                if name_only:
                    app_specs.append(app)

                elif version == 'latest':
                    app_specs.append((app, update or UpdateFreq.DEFAULT.name.lower()))

                elif len(version.split('.')) < 3:
                    app_specs.append((f'{app}=={version}.*', update or UpdateFreq.DEFAULT.name.lower()))

                else:  # Specific version without auto-update
                    app_specs.append((f'{app}=={version}', None))

        return app_specs

    @lru_cache()
    def _pkg_info(self, path=None):
        """ Get scripts and entry points from the app """
        if not path:
            if not self.current_path:
                return

            path = self.current_path

        inspect_py = Path(__file__).parent / 'inspect_app.py'

        try:
            info = run(f"""set -e
                source {path / 'bin/activate'}
                python {inspect_py} {self.name}
                """, executable='/bin/bash', stderr=STDOUT, shell=True)
            return json.loads(info)

        except Exception as e:
            debug('! Can not get package distribution info because: %s', e)

    def uninstall(self):
        """ Uninstall app """
        info('Uninstalling %s', self.name)

        try:
            crontab.remove(self._crontab_id)
        except exceptions.MissingError as e:
            debug('Could not remove crontab for %s: %s', self._crontab_id, e)

        for script in self.scripts():
            script_symlink = self.paths.symlink_root / script
            if ((script_symlink.exists() or script_symlink.is_symlink())
                    and str(script_symlink.resolve()).startswith(str(self.path))):
                script_symlink.unlink()

        shutil.rmtree(self.path)


class AppsPath:
    """
    Checks user access and determine if we are installing to system vs user path.

    System paths are /opt and /usr/local/bin and user paths are in ~
    """
    # System install paths (e.g. root)
    SYSTEM_INSTALL_ROOT = Path('/opt/apps')
    SYSTEM_SYMLINK_ROOT = Path('/usr/local/bin')
    SYSTEM_LOG_ROOT = Path('/var/log/autopip')

    # Local install paths (e.g. user owned /usr/local on macOS)
    _LOCAL_BASE = Path('/usr/local')
    LOCAL_INSTALL_ROOT = _LOCAL_BASE / 'opt' / 'apps'
    LOCAL_SYMLINK_ROOT = _LOCAL_BASE / 'bin'
    LOCAL_LOG_ROOT = _LOCAL_BASE / 'var' / 'log' / 'autopip'

    # User install paths
    USER_INSTALL_ROOT = Path.home() / '.apps'
    USER_SYMLINK_ROOT = Path.home() / 'bin'
    USER_LOG_ROOT = USER_INSTALL_ROOT / '.log'

    def __init__(self):
        #: Root to install apps. This will be set at runtime based on permission by :meth:`_set_roots`
        self.install_root = None

        #: Root to install symlinks. This will be set at runtime based on permission by :meth:`_set_roots`
        self.symlink_root = None

        #: Root to write log files. This will be set at runtime based on permission by :meth:`_set_roots`
        self.log_root = None

        #: Indicates if we are using user paths as we do not have access to system paths.
        self.is_user = False

        self._set_roots()

    def _set_roots(self):
        """ Check to see if we have access to system paths and set roots accordingly. """
        system_reasons = []
        local_reasons = []

        # Check system
        if not os.access(self.SYSTEM_INSTALL_ROOT.parent, os.W_OK):
            system_reasons.append(f'No permission to write to {self.SYSTEM_INSTALL_ROOT.parent}')

        if not os.access(self.SYSTEM_SYMLINK_ROOT, os.W_OK):
            system_reasons.append(f'No permission to write to {self.SYSTEM_SYMLINK_ROOT}')

        if not (os.access(self.SYSTEM_LOG_ROOT.parent, os.W_OK)):
            system_reasons.append(f'No permission to write to {self.SYSTEM_LOG_ROOT.parent}')

        if system_reasons:
            debug('Not using system paths because:\n%s', '* ' + '\n* '.join(system_reasons))

        # Check local
        if system_reasons:
            if not (os.access(self.LOCAL_INSTALL_ROOT.parent, os.W_OK)
                    or not self.LOCAL_INSTALL_ROOT.parent.exists()
                    and os.access(self.LOCAL_INSTALL_ROOT.parent.parent, os.W_OK)):
                local_reasons.append(f'No permission to write to {self.LOCAL_INSTALL_ROOT.parent}')

            if not os.access(self.LOCAL_SYMLINK_ROOT, os.W_OK):
                local_reasons.append(f'No permission to write to {self.LOCAL_SYMLINK_ROOT}')

            if not (os.access(self.LOCAL_LOG_ROOT.parent, os.W_OK)
                    or not self.LOCAL_LOG_ROOT.parent.exists()
                    and os.access(self.LOCAL_LOG_ROOT.parent.parent, os.W_OK)):
                local_reasons.append(f'No permission to write to {self.LOCAL_LOG_ROOT.parent}')

            if local_reasons:
                debug('Not using local paths because:\n%s', '* ' + '\n* '.join(local_reasons))

        if not system_reasons:
            self.install_root = self.SYSTEM_INSTALL_ROOT
            self.symlink_root = self.SYSTEM_SYMLINK_ROOT
            self.log_root = self.SYSTEM_LOG_ROOT

        elif not local_reasons:
            self.install_root = self.LOCAL_INSTALL_ROOT
            self.symlink_root = self.LOCAL_SYMLINK_ROOT
            self.log_root = self.LOCAL_LOG_ROOT

        else:
            self.install_root = self.USER_INSTALL_ROOT
            self.symlink_root = self.USER_SYMLINK_ROOT
            self.log_root = self.USER_LOG_ROOT
            self.is_user = True

        self.install_root.mkdir(parents=True, exist_ok=True)
        self.symlink_root.mkdir(parents=True, exist_ok=True)
        self.log_root.mkdir(parents=True, exist_ok=True)

    def covers(self, path):
        """ True if the given path belongs to autopip """
        path = path.resolve() if isinstance(path, PurePath) else path
        return (str(path).startswith(str(self.SYSTEM_INSTALL_ROOT))
                or str(path).startswith(str(self.LOCAL_INSTALL_ROOT))
                or str(path).startswith(str(self.USER_INSTALL_ROOT)))



import os
import re
from subprocess import CalledProcessError
from time import time

from mock import MagicMock, Mock, call
import pytest

from autopip.utils import run
from autopip.constants import PYTHON_VERSION


def test_autopip_help(autopip, capsys):
    autopip('', raises=SystemExit)
    stdout, stderr = capsys.readouterr()

    assert 'usage: autopip' in stdout


def test_autopip_common(monkeypatch, autopip, capsys, mock_paths):
    system_root, _, _ = mock_paths
    mock_run = MagicMock()
    monkeypatch.setattr('autopip.crontab.run', mock_run)
    monkeypatch.setattr('autopip.crontab.randint', Mock(return_value=10))

    # Install latest
    stdout = autopip('install bumper --update hourly')
    assert 'Installing bumper to' in stdout
    assert 'Updating script symlinks in' in stdout
    assert '+ bump' in stdout
    assert len(stdout.split('\n')) == 5

    assert run([str(system_root / 'bin' / 'bump'), '-h']).startswith('usage: bump')

    assert len(mock_run.call_args_list) == 6
    assert mock_run.call_args_list[0:-1] == [
        call('which crontab', shell=True, stderr=-2),
        call('ps -ef | grep /usr/sbin/cron | grep -v grep', shell=True, stderr=-2),
        call('crontab -l | grep autopip', shell=True, stderr=-2),
        call('which crontab', shell=True, stderr=-2),
        call('ps -ef | grep /usr/sbin/cron | grep -v grep', shell=True, stderr=-2)
        ]
    update_call = re.sub('/tmp/.*/system/', '/tmp/system/',
                         re.sub('/home/.*virtualenvs/autopip[^/]*', '/home/venv/autopip',
                                mock_run.call_args_list[-1][0][0]))
    assert update_call == (
        '( crontab -l | grep -vi "autopip update"; echo "10 * * * * PATH=/usr/local/bin:\$PATH '
        '/home/venv/autopip/bin/autopip update 2>&1 >> /tmp/system/log/cron.log" ) | crontab -')

    assert 'system/bumper/0.1.13' in autopip('list')
    assert autopip('list --scripts').split('\n')[1].strip().endswith('/bin/bump')

    # Already installed
    mock_run.reset_mock()
    assert autopip('install bumper --update hourly') == """\
bumper is up-to-date
Hourly auto-update enabled via cron service
Scripts are in /tmp/system/bin: bump
"""
    assert mock_run.call_count == 6

    # Update manually
    assert autopip('update') == 'bumper is up-to-date\n'
    assert autopip('update blah') == 'No apps found matching: blah\nAvailable apps: bumper\n'

    # Update via cron
    assert autopip('update', isatty=False) == ''
    bumper_root = system_root / 'bumper'
    last_modified = bumper_root.stat().st_mtime
    with monkeypatch.context() as m:
        m.setattr('autopip.manager.time', Mock(return_value=time() + 3600))
        assert autopip('update', isatty=False) == ''
        current_modified = bumper_root.stat().st_mtime
        assert current_modified > last_modified

    # Wait for new version
    mock_sleep = Mock(side_effect=[0, 0, 0, Exception('No new version')])
    monkeypatch.setattr('autopip.manager.sleep', mock_sleep)

    stdout, e = autopip('update bumper --wait', raises=SystemExit)
    assert stdout.startswith('! No new version')

    stdout, _ = capsys.readouterr()
    lines = stdout.split('\n')
    assert len(lines) == 5
    assert lines[0] == 'Waiting for new version of bumper to be published...'.ljust(80)
    assert lines[-2] == '\033[1AWaiting for new version of bumper to be published...'.ljust(80)

    assert mock_sleep.call_count == 4

    # Uninstall
    mock_run.reset_mock()
    assert autopip('uninstall bumper') == 'Uninstalling bumper\n'
    assert mock_run.call_args_list == [
        call('which crontab', shell=True, stderr=-2),
        call('ps -ef | grep /usr/sbin/cron | grep -v grep', shell=True, stderr=-2),
        call('( crontab -l | grep -vi "autopip install \\"bumper[^a-z]*\\"" ) | crontab -', shell=True, stderr=-2),
        call('which crontab', shell=True, stderr=-2),
        call('ps -ef | grep /usr/sbin/cron | grep -v grep', shell=True, stderr=-2),
        call('( crontab -l | grep -vi "autopip" ) | crontab -', shell=True, stderr=-2)
    ]

    assert autopip('list') == 'No apps are installed yet.\n'


def test_update(autopip):
    assert autopip('update') == 'No apps installed yet.\n'


def test_install_lib(autopip):
    stdout, e = autopip('install utils-core', raises=SystemExit)
    assert 'Uninstalling utils-core' in stdout
    assert '! Odd, there are no scripts included in the app' in stdout


def test_install_bad_version(autopip, monkeypatch):
    stdout, _ = autopip('install bumper==100.*', raises=SystemExit)
    assert '! No app version matching bumper==100.*' in stdout
    assert 'Available versions: 0.1.8, 0.1.9, 0.1.10, 0.1.11, 0.1.12' in stdout


def test_install_no_path(autopip, monkeypatch):
    monkeypatch.setenv('PATH', '')
    stdout, _ = autopip('install bumper', raises=SystemExit)

    assert stdout == (f'! python{PYTHON_VERSION} does not exist. '
                      'Please install it first, or ensure its path is in PATH.\n')


def test_install_python_2(autopip, mock_paths):
    system_path, _, _ = mock_paths
    assert autopip('install bumper --python 2.7 --update hourly') == """\
Installing bumper to /tmp/system/bumper/0.1.13
Hourly auto-update enabled via cron service
Updating script symlinks in /tmp/system/bin
+ bump
"""
    version = run([str(system_path / 'bumper' / 'current' / 'bin' / 'python'), '--version'],
                  stderr=-2)
    assert version.startswith('Python 2.7')

    assert run([str(system_path / 'bin' / 'bump'), '-h']).startswith('usage: bump')

    assert autopip('uninstall bumper')


def test_install_nonexisting(autopip):
    stdout, _ = autopip('install this-does-not-exist-blah-blah', raises=SystemExit)
    assert stdout.startswith('! No app version found in http')


def test_install_failed(autopip, monkeypatch, mock_run):
    mock_run.side_effect = Exception('install failed')
    monkeypatch.setattr('autopip.manager.run', mock_run)
    stdout, _ = autopip('install utils-core', raises=SystemExit)
    assert '! install failed' in stdout


def test_install_python2_using_python3_mock(autopip, mock_run):
    mock_run.side_effect = CalledProcessError(1, 'cmd',
                                              output='some-pkg is a builtin module since Python 3'.encode())
    stdout, _ = autopip('install pantsbuild.pants==1.6.0', raises=SystemExit)
    assert stdout.startswith(f"""\
Installing pantsbuild.pants to /tmp/system/pantsbuild.pants/1.6.0
Failed to install using Python {PYTHON_VERSION} venv, let's try using Python 2 virtualenv.
Installing pantsbuild.pants to /tmp/system/pantsbuild.pants/1.6.0
some-pkg is a builtin module since Python 3
! Failed to install using Python 2. If this app requires a different Python version, \
please specify it using --python option.
! Command 'cmd' returned non-zero exit status 1.
""")


@pytest.mark.skipif(not os.environ.get('ALL'), reason='Too slow. Set ALL=1 to run')
def test_install_python2_using_python3(autopip, ):
    assert autopip('install pantsbuild.pants==1.6.0') == f"""\
Installing pantsbuild.pants to /tmp/system/pantsbuild.pants/1.6.0
Failed to install using Python {PYTHON_VERSION} venv, let's try using Python 2 virtualenv.
Installing pantsbuild.pants to /tmp/system/pantsbuild.pants/1.6.0
Auto-update will be disabled since we are pinning to a specific version.
To enable, re-run without pinning to specific version with --update option
Updating script symlinks in /tmp/system/bin
+ pants
"""
    assert autopip('uninstall pantsbuild.pants') == 'Uninstalling pantsbuild.pants\n'


def test_install_autopip(autopip, monkeypatch):
    assert autopip('install autopip==1.4.2 --update hourly') == """\
Installing autopip to /tmp/system/autopip/1.4.2
Auto-update will be disabled since we are pinning to a specific version.
To enable, re-run without pinning to specific version with --update option
Updating script symlinks in /tmp/system/bin
+ app
+ autopip
"""
    with monkeypatch.context() as m:
        remove_cron = Mock(side_effect=Exception('failed'))
        m.setattr('autopip.manager.crontab.remove', remove_cron)
        assert autopip('update', isatty=False) == ''
        remove_cron.assert_called_with('autopip')

    stdout = autopip('list auto --scripts')
    assert re.sub(' +autopip', ' ' * 32 + 'autopip', stdout) == """\
autopip  1.4.2  /tmp/system/autopip/1.4.2  
                /tmp/system/bin/app        
                                autopip    
"""  # noqa

    assert autopip('uninstall autopip') == 'Uninstalling autopip\n'


def test_autopip_group(monkeypatch, autopip):
    mock_run = MagicMock()
    monkeypatch.setattr('autopip.crontab.run', mock_run)
    monkeypatch.setattr('autopip.crontab.randint', Mock(return_value=10))

    def mock_group_specs(self, path=None, name_only=False):
        if self.name == 'developer-tools':
            return ['bumper'] if name_only else [('bumper==0.1.10', None)]
        else:
            return []

    monkeypatch.setattr('autopip.manager.App.group_specs', mock_group_specs)

    # Install latest
    stdout = autopip('install developer-tools --update weekly')
    installed_version = stdout.split('\n')[0].split('/')[-1]
    assert 'Installing developer-tools to' in stdout
    assert 'Updating script symlinks in' in stdout
    assert 'This app has defined "autopip" entry points to install: bumper==0.1.10' in stdout
    assert '+ bump' in stdout
    assert len(stdout.split('\n')) == 7

    assert len(mock_run.call_args_list) == 6
    assert mock_run.call_args_list[0:-1] == [
        call('which crontab', shell=True, stderr=-2),
        call('ps -ef | grep /usr/sbin/cron | grep -v grep', shell=True, stderr=-2),
        call('crontab -l | grep autopip', shell=True, stderr=-2),
        call('which crontab', shell=True, stderr=-2),
        call('ps -ef | grep /usr/sbin/cron | grep -v grep', shell=True, stderr=-2)
        ]
    update_call = re.sub('/tmp/.*/system/', '/tmp/system/',
                         re.sub('/home/.*virtualenvs/autopip[^/]*', '/home/venv/autopip',
                                mock_run.call_args_list[-1][0][0]))
    assert update_call == (
        '( crontab -l | grep -vi "autopip update"; echo "10 * * * * PATH=/usr/local/bin:\$PATH '
        '/home/venv/autopip/bin/autopip update 2>&1 >> /tmp/system/log/cron.log" ) | crontab -')

    assert 'system/bumper/0.1.10' in autopip('list')
    assert f'system/developer-tools/{installed_version}' in autopip('list')
    assert autopip('list --scripts').split('\n')[1].strip().endswith('/bin/bump')
    assert autopip('list blah') == 'No apps matching "blah"\n'

    # Uninstall autopip
    assert autopip('uninstall autopip') == ('! autopip can not be uninstalled until other apps are uninstalled: '
                                            'bumper developer-tools\n')

    # Update
    assert autopip('update') == """\
bumper is up-to-date [per spec: ==0.1.10]
developer-tools is up-to-date
This app has defined "autopip" entry points to install: bumper==0.1.10
"""

    # Already installed
    mock_run.reset_mock()
    assert autopip(f'install developer-tools=={installed_version}') == """\
developer-tools is up-to-date [per spec: ==1.0.7]
Auto-update will be disabled since we are pinning to a specific version.
To enable, re-run without pinning to specific version with --update option
This app has defined "autopip" entry points to install: bumper==0.1.10
bumper is up-to-date [per spec: ==0.1.10]
Scripts are in /tmp/system/bin: bump
"""
    assert mock_run.call_args_list == [
        call('which crontab', shell=True, stderr=-2),
        call('ps -ef | grep /usr/sbin/cron | grep -v grep', shell=True, stderr=-2),
        call('( crontab -l | grep -vi "autopip install \\"developer-tools[^a-z]*\\"" ) | crontab -',
             shell=True, stderr=-2)
    ]

    # Uninstall
    mock_run.reset_mock()
    assert autopip('uninstall developer-tools') == """\
Uninstalling developer-tools
This app has defined "autopip" entry points to uninstall: bumper
Uninstalling bumper
"""
    assert mock_run.call_args_list == [
        call('which crontab', shell=True, stderr=-2),
        call('ps -ef | grep /usr/sbin/cron | grep -v grep', shell=True, stderr=-2),
        call('( crontab -l | grep -vi "autopip install \\"developer-tools[^a-z]*\\"" ) | crontab -',
             shell=True, stderr=-2),
        call('which crontab', shell=True, stderr=-2),
        call('ps -ef | grep /usr/sbin/cron | grep -v grep', shell=True, stderr=-2),
        call('( crontab -l | grep -vi "autopip install \\"bumper[^a-z]*\\"" ) | crontab -',
             shell=True, stderr=-2),
        call('which crontab', shell=True, stderr=-2),
        call('ps -ef | grep /usr/sbin/cron | grep -v grep', shell=True, stderr=-2),
        call('( crontab -l | grep -vi "autopip" ) | crontab -', shell=True, stderr=-2)
    ]

    assert autopip('list') == 'No apps are installed yet.\n'



from logging import debug
import re
from subprocess import check_output


def run(*args, **kwargs):
    debug('Running: %s', args[0])
    return check_output(*args, **kwargs).decode('utf-8')


def sorted_versions(versions):
    version_sep_re = re.compile('[^0-9]+')
    return sorted(versions, key=lambda v: tuple(map(int, version_sep_re.split(v))))



from pathlib import Path
import subprocess

import pytest

from autopip.constants import PYTHON_VERSION


@pytest.mark.parametrize('python', ['python2', 'python3'])
def test_check(python):
    check_script = str(Path(__file__).parent.parent / 'etc' / 'check-python.py')
    cmd = [python, check_script]

    if not(PYTHON_VERSION == '3.6' and python == 'python2'):
        cmd.extend(['--version', PYTHON_VERSION])

    subprocess.check_call(cmd)



class MissingError(Exception):
    """ Indicates something is missing """


class FailedAction(Exception):
    """ Indicates a specific action failed """


class InvalidAction(Exception):
    """ Indicates a specific action failed """
